/* Copyright (C) 2010, Alexander Graf <agraf@znc.in>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include "lib.h"
#include "pab.h"
#include "path.h"

#include "ext2fs.h"

/* information stored with each instance */
struct id {
	struct ext2_super_block sb;
	int groups;
	struct ext2_group_desc *gd;
	struct ext2_inode ino;
};

/* maximum steps in a blockpath */
#define EXT2_BLOCKPATH	4

static int ext2fs_get_inode(struct p_instance *inst, int nr,
			    struct ext2_inode *ino)
{
	struct id *id = inst->data;
	int gnr = (nr-1) / id->sb.s_inodes_per_group;
	int inr = (nr-1) % id->sb.s_inodes_per_group;
	int iaddr;

	iaddr  = id->gd[gnr].bg_inode_table * EXT2_BLOCK_SIZE(id->sb);
	iaddr += inr * EXT2_INODE_SIZE(id->sb);

	return p_read(inst->parent, ino, iaddr, sizeof(struct ext2_inode));
}

static int ext2fs_get_blockpath(struct p_instance *inst, unsigned int block,
				int *path)
{
	struct id *id = inst->data;

	if (block < EXT2_NDIR_BLOCKS) {
		/* direct block */
		*path = block;
		return 1;
	} else if ((block -= EXT2_NDIR_BLOCKS) <
		   EXT2_ADDR_PER_BLOCK(id->sb)) {
		/* indirect block */
		*path++ = EXT2_IND_BLOCK;
		*path   = block;
		return 2;
	} else if ((block -= EXT2_ADDR_PER_BLOCK(id->sb)) <
		   square(EXT2_ADDR_PER_BLOCK(id->sb))) {
		/* double indirect block */
		*path++ = EXT2_DIND_BLOCK;
		*path++ = block / EXT2_ADDR_PER_BLOCK(id->sb);
		*path   = block % EXT2_ADDR_PER_BLOCK(id->sb);
		return 3;
	} else {
		/* triple indirect block */
		block -= square(EXT2_ADDR_PER_BLOCK(id->sb));
		*path++ = EXT2_TIND_BLOCK;
		*path++ = block / square(EXT2_ADDR_PER_BLOCK(id->sb));
		block %= square(EXT2_ADDR_PER_BLOCK(id->sb));
		*path++ = block / EXT2_ADDR_PER_BLOCK(id->sb);
		*path   = block % EXT2_ADDR_PER_BLOCK(id->sb);
		return 4;
	}
}

static int ext2fs_get_block(struct p_instance *inst, struct ext2_inode *ino,
			    int block)
{
	struct id *id = inst->data;
	int addr;
	int blockpath[EXT2_BLOCKPATH];
	int bpl;
	int i;
	int r;
	unsigned int pos;

	bpl = ext2fs_get_blockpath(inst, block, blockpath);

	addr = ino->i_block[blockpath[0]];

	for (i = 1; i < bpl; i++) {
		pos  = addr * EXT2_BLOCK_SIZE(id->sb);
		pos += blockpath[i] * EXT2_ADDR_SIZE;
		r = p_read(inst->parent, &addr, pos, EXT2_ADDR_SIZE);
		if (r < 0)
			return r;
	}

	return addr;
}

static int ext2fs_read_data(struct p_instance *inst, struct ext2_inode *ino,
			    unsigned int addr, void *buf, unsigned int count)
{
	struct id *id = inst->data;
	int block;
	int start;
	unsigned int size;
	int rv = 0;
	int r;

	if (addr + count > ino->i_size)
		count = ino->i_size - addr;
	if (count <= 0)
		return 0;

	while (count) {
		block = ext2fs_get_block(inst, ino,
					 addr / EXT2_BLOCK_SIZE(id->sb));
		if (block < 0)
			return block;

		start = (block * EXT2_BLOCK_SIZE(id->sb)) +
			(addr % EXT2_BLOCK_SIZE(id->sb));

		size = EXT2_BLOCK_SIZE(id->sb) -
			(addr % EXT2_BLOCK_SIZE(id->sb));
		size = min(size, count);

		r = p_read(inst->parent, buf + rv, start, size);
		if (r < 0)
			return r;

		addr  += size;
		rv    += size;
		count -= size;
	}

	return rv;
}

#define CD(de, s, l)	(de.name_len == l && strncmp(de.name, s, l) == 0)

static int ext2fs_search(struct p_instance *inst, const char *path)
{
	int ino;	/* current inode */
	struct ext2_inode in;
	struct ext2_dir_entry de;
	const char *p;	/* path - already processed parts skipped */
	int ctns;	/* chars to next slash */
	int last;	/* 1 if looking for last part of pathname */
	int addr;	/* address in directory */
	void *tmp;
	int r;

	ino = EXT2_ROOT_INO;
	p = path;
	last = 0;

search:
	tmp = strchr(p, '/');
	if (!tmp) {
		last = 1;
		ctns = strlen(p);
	} else {
		ctns = (unsigned int) tmp - (unsigned int) p;
	}
	if (!ctns)
		return ino;

	r = ext2fs_get_inode(inst, ino, &in);
	if (r < 0)
		return r;

	/* TODO: not fast */

	addr = 0;
	while ((r = ext2fs_read_data(inst, &in, addr, &de,
				     sizeof(struct ext2_dir_entry))) > 0) {
		addr += de.rec_len;

		if (!CD(de, p, ctns))
			continue;
		if (last)
			return de.inode;
		if (de.file_type != EXT2_FT_DIR) {
			errmsg = "ext2: Not a directory.";
			return -1;
		}

		ino = de.inode;
		p += ctns + 1;
		goto search;
	}

	if (r < 0)
		return r;

	errmsg = "ext2: File not found.";
	return -1;
}

static int ext2fs_setup(struct p_instance *inst, char *arg)
{
	int r;
	struct id *id = 0;

	if (!inst->parent) {
		errmsg = "ext2: No parent given.";
		return 0;
	}

	if (!arg) {
		errmsg = "ext2: No argument given.";
		return 0;
	}

	/*
	 * From Armux' ext2fs_mount()
	 */

	id = inst->data = malloc(sizeof(struct id));
	id->gd = 0;

	/* read superblock.  First block is reserved for bootloader, second
	 * block is superblock. */
	r = p_read(inst->parent, &id->sb, 1024,
		   sizeof(struct ext2_super_block));
	if (r < 0)
		goto err;

	if (id->sb.s_magic != EXT2_SB_MAGIC) {
		errmsg = "ext2: Invalid superblock magic.";
		goto err;
	}

	id->groups = blocks(id->sb.s_blocks_count, id->sb.s_blocks_per_group);
	id->gd = malloc(id->groups * sizeof(struct ext2_group_desc));

	/* read group descriptors.  They are in the block behind the superblock
	 * (third block) */
	r = p_read(inst->parent, id->gd, 2048,
		   id->groups * sizeof(struct ext2_group_desc));
	if (r < 0)
		goto err;

	/*
	 * From Armux' ext2fs_open()
	 */

	r = ext2fs_search(inst, arg);
	if (r < 0)
		goto err;

	r = ext2fs_get_inode(inst, r, &id->ino);
	if (r < 0)
		goto err;

	return 1;

err:
	if (id) {
		if (id->gd)
			free(id->gd);
		free(id);
	}
	return 0;
}

static void ext2fs_remove(struct p_instance *inst)
{
	struct id *id = inst->data;

	free(id->gd);
	free(id);
}

static int ext2fs_read(struct p_instance *inst, void *buf,
		       unsigned int off, unsigned int count)
{
	struct id *id = inst->data;

	return ext2fs_read_data(inst, &id->ino, off, buf, count);
}

static const struct p_handles ext2fs_handles = {
	.setup  = ext2fs_setup,
	.remove = ext2fs_remove,
	.read   = ext2fs_read
};

static void ext2fs_init()
{
	p_regmod(&ext2fs_handles, "ext2", "Second extended file system");
}
INIT(ext2fs_init);
