/* Copyright (C) 2008-2010, Alexander Graf <agraf@znc.in>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include "cpu.h"
#include "dev.h"
#include "lib.h"

/*
 * stdio
 */

/* makros for printf */
#define CHAR_CPY(c)				\
	dest[rv++] = c;
#define FILL_CPY(x)				\
	for (i = minlen - (x); i > 0; i--)	\
		CHAR_CPY(fillc)
#define STR_CPY(s)				\
	i = 0;					\
	while (s[i])				\
		CHAR_CPY(s[i++]);

static void itos(unsigned long id, char *s, int base, int is_signed)
{
	static const char chars[] = "0123456789abcdef";
	unsigned char j = 0;
	int negative = 0;
	char *p1, *p2, tmp;

	if (is_signed && (signed) id < 0) {
		negative = 1;
		id = -(signed) id;
	}

	do {
		s[j++] = chars[id % base];
		id /= base;
	} while (id != 0);

	if (negative)
		s[j++] = '-';

	s[j] = 0;

	/* reverse the buf (we have written in wrong order) */
	p1 = s;
	p2 = &s[j - 1];
	while (p1 < p2) {
		tmp = *p1;
		*p1 = *p2;
		*p2 = tmp;
		p1++;
		p2--;
	}
}

/* TODO: read the TODOs inside this function.  Possibly the best is a complete
 * redesign of it (or copy it from somewhere). */
int vsprintf(char *dest, const char *format, va_list ap)
{
	int rv = 0, i;
	char c;
	char buf[12];
	const char *s;

	int minlen, alt, leftalign;
	char fillc;

	/* TODO: make this POSIX conform, smaller and vsnprintf */

	while ((c = *format++) != 0) {
		/* variable references begin with '%' */
		if (c != '%') {
			CHAR_CPY(c);
			continue;
		}

		/* reset to default config */
		minlen = 0;
		fillc = ' ';
		alt = 0;
		leftalign = 0;

		c = *format++;

		/* TODO: don't care about range of indicators like #, 0, - */

		/* # specifies the "alternative format", e.g. 0x before
		 * hexadecimal numbers. */
		if (c == '#') {
			alt = 1;
			c = *format++;
		}

		/* - specifies that the argument content should be aligned on
		 * the left side instead of the right */
		if (c == '-') {
			leftalign = 1;
			c = *format++;
		}

		/* may be followed by '0' to tell that they should be
		 * right-aligned by filling with '0' instead of ' ' */
		if (c == '0') {
			fillc = c;
			c = *format++;
		}

		/* may be followed by the minimum field size */
		while ('0' <= c && c <= '9') {
			minlen *= 10;
			minlen += c - '0';
			c = *format++;
		}

		/* must be followed by type indicator */
		switch (c) {
			case 'c':
				CHAR_CPY((char) va_arg(ap, int));
				break;
			case 'o':
				/* TODO: leftalign not implemented */
				itos(va_arg(ap, unsigned), buf, 8, 0);
				if (alt) {
					if (fillc == ' ')
						FILL_CPY(strlen(buf) + 1);
					STR_CPY("0");
					if (fillc == '0')
						FILL_CPY(strlen(buf) + 1);
				} else
					FILL_CPY(strlen(buf));
				STR_CPY(buf);
				break;
			case 'd':
			case 'i':
				itos(va_arg(ap, int), buf, 10, 1);
				if (leftalign) {
					STR_CPY(buf);
					FILL_CPY(strlen(buf));
				} else {
					FILL_CPY(strlen(buf));
					STR_CPY(buf);
				}
				break;
			case 'x':
				/* TODO: leftalign not implemented */
				itos(va_arg(ap, unsigned), buf, 16, 0);
				if (alt) {
					if (fillc == ' ')
						FILL_CPY(strlen(buf) + 2);
					STR_CPY("0x");
					if (fillc == '0')
						FILL_CPY(strlen(buf) + 2);
				} else
					FILL_CPY(strlen(buf));
				STR_CPY(buf);
				break;
			case 's':
				s = va_arg(ap, const char *);
				if (!s)
					s = "(null)";
				if (leftalign) {
					STR_CPY(s);
					FILL_CPY(strlen(s));
				} else {
					FILL_CPY(strlen(s));
					STR_CPY(s);
				}
				break;
			case '%':
				CHAR_CPY('%');
				break;
			default:
				/* we must abort here, because else there would
				 * be the possibility to read behind the string
				 * on fake situations like %#\0 */
				goto end;
		}
	}

end:
	dest[rv] = 0;

	return rv;
}
int sprintf(char *buf, const char *fmt, ...)
{
	int rv;
	va_list ap;

	va_start(ap, fmt);
	rv = vsprintf(buf, fmt, ap);
	va_end(ap);

	return rv;
}

/* printf' buffer size */
#define PRINTF_BUF	128

int vprintf(const char *fmt, va_list ap)
{
	char buf[PRINTF_BUF];
	int rv, i;

	rv = vsprintf(buf, fmt, ap);

	for (i = 0; i < rv; i++)
		uart_putchar(buf[i]);

	return rv;
}

int printf(const char *fmt, ...)
{
	int rv;
	va_list ap;

	va_start(ap, fmt);
	rv = vprintf(fmt, ap);
	va_end(ap);

	return rv;
}

int puts(const char *s)
{
	while (*s)
		uart_putchar(*s++);
	uart_putchar('\n');
	return 0;
}

/*
 * string
 */

void *memcpy(void *s1, const void *s2, unsigned int n)
{
	char *d = s1;
	const char *s = s2;
	int *r;
	const int *q;

	if (n < 16)
		/* too much overhead for using wordwise memcpy */
		goto bytewise;

	if (((int) s1 & 3) != ((int) s2 & 3))
		/* pointers cannot be both aligned to word boundaries */
		goto bytewise;

	/* align on word boundary */
	while (((int) d & 3) != 0) {
		*d++ = *s++;
		n--;
	}

	/* wordwise copying till no whole word is left */
	r = (int *) d;
	q = (const int *) s;
	while (n >= 4) {
		*r++ = *q++;
		n -= 4;
	}

	/* do bytewise what is missing */
	d = (char *) r;
	s = (const char *) q;
bytewise:
	while (n--)
		*d++ = *s++;

	return s1;
}

void *memset(void *s, int c, unsigned int n)
{
	char *q = s;
	int *r;
	int word;
	char ch = (char) c;

	if (n < 16)
		/* too much overhead for using wordwise memset */
		goto bytewise;

	/* build up word, or use 0 if c is 0 (happens often) */
	if ((word = ch) != 0) {
		word |= word << 8;
		word |= word << 16;
	}

	/* align on word boundary */
	while (((int) q & 3) != 0) {
		*q++ = ch;
		n--;
	}

	/* wordwise copying till no whole word is left */
	r = (int *) q;
	while (n >= 4) {
		*r++ = word;
		n -= 4;
	}

	/* do bytewise what is missing */
	q = (char *) r;
bytewise:
	while (n--)
		*q++ = ch;

	return s;
}

int strcmp(const char *s1, const char *s2)
{
	int c = 0;

	while (!(c = *s1++ - *s2++) && (*s1 || *s2)) ;

	return c;
}

int strncmp(const char *s1, const char *s2, unsigned int n)
{
	int c = 0;

	while (n-- && !(c = *s1++ - *s2++) && (*s1 || *s2)) ;

	return c;
}

char *strchr(const char *s, int c)
{
	while (*s != (char) c) {
		if (!*s)
			return 0;
		s++;
	}

	return (char *) (int) s;
}

unsigned int strlen(const char *s)
{
	int i = 0;

	while (s[i++] != 0) ;

	return i - 1;
}

/*
 * stdlib
 */

#define BLOCKSIZE	4			/* useful, since structs are 32
						 * bit aligned and padded */
#define BPE		(RAM_SIZE / BLOCKSIZE)	/* blocks per mpg entry */
#define BITMAPSIZE	(BPE / 32)		/* number of ints of bitmap */

static int mpg_bitmap[BITMAPSIZE];

void *malloc(unsigned int size)
{
	int i, j;
	int blk;
	void *rv;

	if (!size)
		/* save time & space */
		return 0;

	size += 4;	/* needing an int to store length of allocation */

	blk = blocks(size, BLOCKSIZE);

	/* scan through free blocks */
	for (i = 0; i + blk < BPE; i++) {
		if (mpg_bitmap[i / 32] & 1 << (i % 32))
			/* block used */
			continue;

		/* check whether there are enough continious free
		 * blocks */
		for (j = i + 1; j < i + blk; j++)
			if (mpg_bitmap[j / 32] & 1 << (j % 32))
				break;

		if (j < i + blk) {
			/* not enough free blocks */
			i = (j - 1);
			continue;
		}

		break;
	}

	if (i + blk >= BPE) {
		printf("No memory left.\n");
		/* TODO: reset (nice) */
		for (;;) ;
	}

	/* mark blocks as used */
	for (j = i; j < blk + i; j++)
		mpg_bitmap[j / 32] |= 1 << (j % 32);

	/* get address of freshly allocated allocation */
	rv = (void *) (RAM_START + i * BLOCKSIZE);

	/* save number of blocks */
	*(int *) rv = blk;

	/* return */
	return rv + 4;
}

void free(void *ptr)
{
	int blkcnt, blknr;
	int i;

	if (!ptr)
		/* malloc(0) returns 0, so this is valid */
		return;

	ptr -= 4;

	blkcnt = *(int *) ptr;

	/* get block number in this mpg */
	blknr = ((unsigned long) ptr - RAM_START) / BLOCKSIZE;

	/* mark blocks as free */
	for (i = blknr; i < blknr + blkcnt; i++)
		mpg_bitmap[i / 32] &= ~(1 << (i % 32));
}

extern void __bss_end__;
#define STACKSIZE	4096

void mm_init()
{
	int i, blkcnt, blknr;

	/* mark .bss blocks as used */
	blkcnt = blocks((unsigned int) &__bss_end__ - RAM_START, BLOCKSIZE);
	blknr = 0;
	for (i = blknr; i < blknr + blkcnt; i++)
		mpg_bitmap[i / 32] |= 1 << (i % 32);

	/* mark stack blocks as used */
	blkcnt = blocks(STACKSIZE, BLOCKSIZE);
	blknr = blocks(RAM_SIZE, BLOCKSIZE) - blkcnt;
	for (i = blknr; i < blknr + blkcnt; i++)
		mpg_bitmap[i / 32] |= 1 << (i % 32);
}
