/* Copyright (C) 2010, Alexander Graf <agraf@znc.in>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include "dev.h"
#include "lib.h"
#include "pab.h"
#include "path.h"

const char *errmsg;

static void cmd_cat(char *arg);
static void cmd_help(char *arg);
static void cmd_reset(char *arg);
static void cmd_source(char *arg);

static const struct commands {
	void (*func)(char *arg);/* function pointer */
	const char *name;	/* name of command */
	const char *arg;	/* brief argument description or NULL */
	const char *desc;	/* brief description printed with "help" */
} commands[] = {
	{ cmd_cat,   "cat",   "path",	"Show contents of a file" },
	{ cmd_help,  "help",  0,	"Show usage information" },
	{ cmd_load,  "load",  "path [arg]", "Load and execute ELF file" },
	{ cmd_reset, "reset", 0,	"Reboot device (software reset)" },
	{ cmd_source,"source", "path",	"Execute commands from file" },
	{ 0, 0, 0, 0 }
};

static void cmd_cat(char *arg)
{
	char buf[512];
	struct p_instance *inst;
	int i, r, c = 0;

	if (!arg) {
		puts("No argument given.");
		return;
	}

	inst = p_open(arg);
	if (!inst) {
		puts(errmsg);
		return;
	}

	do {
		r = p_read(inst, buf, c, sizeof(buf));
		c += r;
		for (i = 0; i < r; i++)
			uart_putchar(buf[i]);
	} while (r == sizeof(buf));

	if (r == -1) {
		puts(errmsg);
		return;
	}

	p_remove(inst);
}

static void cmd_help(char *arg)
{
	int i;
	static const char *pathinfo =
"A path defines a hierarchy of modules, directories and a file, i.e.\n"
"\"mod1:arg1[|mod2:arg2[|...]]|fsmod:path/to/file\".  Modules are drivers for\n"
"devices or on-device structures such as partitions or filesystems.  E.g. a\n"
"valid path could be \"mmc|part:0|ext2:boot/kernel.elf\"";

	(void) arg;

	puts("Implemented commands:");
	for (i = 0; commands[i].func; i++)
		printf(" %-7s %-11s %s\n",
		       commands[i].name,
		       commands[i].arg ? commands[i].arg : "",
		       commands[i].desc);

	puts("\nImplemented modules:");
	p_listmods();

	puts("");
	puts(pathinfo);
}

static void cmd_reset(char *arg)
{
	(void) arg;
	asm volatile ("mov pc, #0x0");
}

static void execcommand(char *buf)
{
	char *arg;
	int i;

	/* arg should point to the argument of the command, or to 0, if the
	 * command has no argument. */
	arg = strchr(buf, ' ');
	if (arg) {
		*arg = 0;	/* buf should point to the command name */
		arg++;
	}

	/* call command handle */
	for (i = 0; commands[i].func; i++)
		if (!strcmp(buf, commands[i].name)) {
			commands[i].func(arg);
			return;
		}

	printf("Unknown command, \"%s\".\n", buf);
}

static void cmd_source(char *arg)
{
	char buf[512];
	struct p_instance *inst;
	int r, c = 0;
	char *cmd, *nextcmd;

	if (!arg) {
		puts("No argument given.");
		return;
	}

	inst = p_open(arg);
	if (!inst) {
		puts(errmsg);
		return;
	}

	do {
		r = p_read(inst, buf, c, sizeof(buf) - 1);
		if (r < 0)
			break;

		buf[r] = 0;

		cmd = buf;
		do {
			nextcmd = strchr(cmd, '\n');
			if (nextcmd) {
				*nextcmd = 0;
				nextcmd++;
				c += nextcmd - cmd;
			}

			/* remove preceding spaces */
			while (*cmd == ' ' || *cmd == '\t')
				cmd++;

			if (!*cmd || *cmd == '#') ; /* empty line / comment */
			else if (*cmd == '"')
				/* special command to display something */
				puts(cmd);
			else {	/* real command */
				puts(cmd);
				execcommand(cmd);
			}

			cmd = nextcmd;
		} while (cmd);
	} while (r == sizeof(buf) - 1);

	if (r == -1) {
		puts(errmsg);
		return;
	}

	p_remove(inst);
}

/* Read from UART till one command is there. Then execute it. */
static void prompt()
{
	char *buf;

	printf(">>> ");

	/* fill the buffer with the input */
	buf = uart_getline();

	if (*buf == 0)
		/* empty line.  just skip command processing so that we just
		 * prompt again. */
		return;

	execcommand(buf);
}

/* call all functions inserted in the linker-created table using INIT() */
extern void initfuncs_start, initfuncs_end;
static void init_per_table()
{
	unsigned long *p;
	void (*f)();

	for (p = &initfuncs_start;
	     (unsigned long) p < (unsigned long) &initfuncs_end;
	     p++) {
		f = (void *) *(unsigned long *) p;
		f();
	}
}

void init()
{
#ifdef INITCMD
	char *buf;
#endif

	uart_init();
	mm_init();
	printf("\npab bootloader version " VERSION ", ");
	printf("copyright (C) 2010, Alexander Graf.\n");

	init_per_table();

#ifdef INITCMD
	printf("%s (builtin)\n", INITCMD);
	/* copying to kill const specifier */
	buf = alloca(sizeof(INITCMD));
	memcpy(buf, INITCMD, sizeof(INITCMD));
	execcommand(buf);
#endif

	printf("If you do not know what to do, type \"help\".\n");
	while (1)
		prompt();
}
