/* Copyright (C) 2008-2010, Alexander Graf <agraf@znc.in>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include "cpu.h"
#include "lib.h"
#include "mmc.h"
#include "pab.h"
#include "path.h"

/* TODO: have generic, more flexible SPI implementation */

#define HIGH_SPEED_MODE	0	/* full speed */
#define LOW_SPEED_MODE	1	/* speed lowered with waiting loops */

/* maximum bits per second in low speed mode */
#define LOW_SPEED_BPS	(150 << 10)

/* wait cycles for low speed mode
 * is divided by 4 because there are 2 waiting loops, and at least 2
 * instructions are needed for such a loop */
#define LOW_SPEED_WCYC	(CPU_CLOCK / LOW_SPEED_BPS / 4)

static void sspi_write(unsigned char byte, int mode)
{
	volatile int t;
	int i;

	for (i = 7; i >= 0; i--) {
		if (mode == LOW_SPEED_MODE)
			for (t = LOW_SPEED_WCYC; t; t--) ;

		if (byte & (1 << i))
			IOSET = DI;
		else
			IOCLR = DI;

		IOCLR = CLK;

		if (mode == LOW_SPEED_MODE)
			for (t = LOW_SPEED_WCYC; t; t--) ;

		IOSET = CLK;
	}

	IOSET = DI;
}

static unsigned char sspi_read(int mode)
{
	volatile int t;
	int i;
	unsigned char rv = 0;

	/* since we always keep DI on high when we do nothing with it, this
	 * function also sends the 0xff dummy byte. */

	for (i = 7; i >= 0; i--) {
		if (mode == LOW_SPEED_MODE)
			for (t = LOW_SPEED_WCYC; t; t--) ;

		IOCLR = CLK;

		if (mode == LOW_SPEED_MODE)
			for (t = LOW_SPEED_WCYC; t; t--) ;

		if (IOPIN & DO)
			rv |= (1 << i);
		IOSET = CLK;
	}

	return rv;
}

/* MMC commands */
#define GO_IDLE_STATE		0	/* resets the MMC */
#define SEND_OP_COND		1	/* activates the MMC' initalization process */
#define READ_SINGLE_BLOCK	17	/* reads a single block */
#define WRITE_BLOCK		24	/* writes a block */

#define BLOCKSIZE	512

/* number of times, something should be retried. */
#define TIMEOUT	200

/* TODO: have generic blocklayer with buffer/cache support */

/* this function sends a command and leaves CS active after it. It returns the
 * answer or -EIO when the card times out */
static int mmc_command(int cmd, int argument, int mode)
{
	unsigned char instruction[6];
	unsigned char rv;
	int i;
	int timeout;

	/* the instructions have the following layout:
	 * 01, 6 bit command, 32 bit argument, 7 bit CRC, 1
	 *
	 * In SPI mode, the CRC needs not to be valid. Nevertheless, we have to
	 * send it when we sending the GO_IDLE_STATE command, since we are in
	 * MMC-Interface mode, when we do so (It's the command which switches to
	 * SPI). */
	instruction[0] = 0x40 | cmd;
	instruction[1] = (argument & 0xff000000) >> 24;
	instruction[2] = (argument & 0x00ff0000) >> 16;
	instruction[3] = (argument & 0x0000ff00) >> 8;
	instruction[4] = (argument & 0x000000ff);
	instruction[5] = (cmd == GO_IDLE_STATE) ? 0x95 : 0xff;

	/* send a dummy byte */
	IOSET = CS;
	sspi_write(0xff, mode);

	/* send the instruction */
	IOCLR = CS;
	for (i = 0; i < 6; i++)
		sspi_write(instruction[i], mode);

	/* wait for a valid answer */
	for (rv = 0xff, timeout = TIMEOUT;
	     rv == 0xff && timeout;
	     rv = sspi_read(mode), timeout--) ;
	if (!timeout) {
		errmsg = "mmc: timed out.";
		return -1;
	}

	return rv;
}

/* reads a block */
static int mmc_read_block(unsigned int start, unsigned int count, void *buf)
{
	unsigned int i;
	int timeout;
	int r;
	unsigned int block = start / BLOCKSIZE;
	unsigned int off   = start % BLOCKSIZE;
	char c;

	/* TODO: do not use READ_SINGLE_BLOCK.  This could avoid the need of
	 * mmc_read() function splitting into block chunks.  */

	if ((r = mmc_command(READ_SINGLE_BLOCK, block * BLOCKSIZE,
			     HIGH_SPEED_MODE)) < 0)
		return 0;

	if (r != 0) {
		/* card sent error code
		 * TODO: evaluate it and printf it */
		errmsg = "mmc: received error code.";
		return 0;
	}

	/* wait for start byte */
	for (timeout = TIMEOUT;
	     sspi_read(HIGH_SPEED_MODE) != 0xfe && timeout;
	     timeout--) ;
	if (!timeout) {
		errmsg = "mmc: timed out.";
		return 0;
	}

	for (i = 0; i < BLOCKSIZE; i++) {
		c = sspi_read(HIGH_SPEED_MODE);
		if (off <= i && i < off + count)
			*((char *) buf++) = c;
	}

	/* ignore 16 bit CRC sum */
	sspi_read(HIGH_SPEED_MODE);
	sspi_read(HIGH_SPEED_MODE);

	IOSET = CS;

	return count;
}

static int mmc_read(struct p_instance *inst, void *buf,
		    unsigned int off, unsigned int count)
{
	int r;
	unsigned int cc;
	unsigned int j = 0, pos = off;

	(void) inst;

	while (j < count) {
		/* Three cases must be considered for setting cc:
		 * 1. current addr might be behind the beginning of a block
		 * 2. current addr might be at the beginning of a block
		 * 3. current addr might be in the last requested block */
		cc = BLOCKSIZE - pos % BLOCKSIZE;	/* 1 */
		if (!cc)				/* 2 */
			cc = BLOCKSIZE;
		if (pos + cc > off + count)		/* 3 */
			cc = off + count - pos;

		if ((r = mmc_read_block(pos, cc, buf + j)) == 0)
			return r;

		pos += r;
		j   += r;
	}

	return j;
}

static const struct p_handles mmc_handles = {
	.setup  = 0,
	.remove = 0,
	.read   = mmc_read
};

static void mmc_init()
{
	int i;
	int timeout;

	if (!mmc_card_inserted()) {
		printf("mmc: no card inserted\n");
		return;
	}

	/* define all output pins (= MMC' input pins) as output pins and set
	 * them to high */
	IODIR |= CS | DI | CLK;
	IOSET = CS | DI | CLK;

	/* send some SPI dummy bytes to the card */
	for (i = 0; i < 15; i++)
		sspi_write(0xff, LOW_SPEED_MODE);

	/* GO_IDLE_STATE is used to reset the MMC and to tell it, that we hate
	 * the MMC interface and would like to use SPI */
	for (timeout = TIMEOUT;
	     mmc_command(GO_IDLE_STATE, 0, LOW_SPEED_MODE) != 1 && timeout;
	     timeout--) ;
	IOSET = CS;
	if (!timeout) {
		puts("mmc: timed out.");
		return;
	}

	/* Tell the card to initalize itself */
	for (timeout = TIMEOUT;
	     mmc_command(SEND_OP_COND, 0, LOW_SPEED_MODE) != 0 && timeout;
	     timeout--) ;
	IOSET = CS;
	if (!timeout) {
		puts("mmc: timed out.");
		return;
	}

	/* TODO: read CSD */
	puts("mmc: initalized.");

	p_regmod(&mmc_handles, "mmc", "MMC access through software-SPI (GPIO)");
}
INIT(mmc_init);
