/* Copyright (C) 2009-2010, Alexander Graf <agraf@znc.in>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include "lib.h"
#include "pab.h"
#include "path.h"

/* format of partition entry */
struct part {
	unsigned char bootind;		/* bit 7 indicates whether bootable */
	unsigned char start_chs[3];	/* start cylinder/head/sector */
	unsigned char sysind;		/* type of partition */
	unsigned char end_chs[3];	/* end cylinder/head/sector */
	unsigned int start;		/* number of first sector */
	unsigned int size;		/* partition size in sectors */
} __attribute__((packed));

/* format of partition table */
struct pt {
	struct part part[4];
	unsigned short magic;
} __attribute__((packed));

/* address of partition table */
#define PT_ADDR		0x1be

/* magic */
#define PT_MAGIC	0xaa55

/* bootable flag of part::bootind */
#define PP_BOOTABLE	(1 << 7)

struct part_ud {
	unsigned int start;
};

static int part_setup(struct p_instance *inst, char *arg)
{
	struct pt pt;
	int partnr;
	int sysind;
	struct part_ud *ud;

	if (!inst->parent) {
		errmsg = "part: No parent given.";
		return 0;
	}

	if (!arg) {
		errmsg = "part: No argument given.";
		return 0;
	}

	if (p_read(inst->parent, &pt, PT_ADDR, sizeof(pt)) == -1)
		return 0;

	if (pt.magic != PT_MAGIC) {
		errmsg = "part: Partition table invalid.";
		return 0;
	}

	if ('0' <= *arg && *arg <= '3')
		/* partition selection by number */
		partnr = *arg - '0';
	else if (*arg == 'b') {
		/* use first partition marked as bootable */
		for (partnr = 0; partnr < 4; partnr++)
			if (pt.part[partnr].bootind & PP_BOOTABLE)
				break;
		if (partnr == 4) {
			errmsg = "part: Partition not found.";
			return 0;
		}
	} else if (*arg == 's') {
		/* partition selection by sysind */
		sysind = strtol(&arg[1], 0, 16);
		for (partnr = 0; partnr < 4; partnr++)
			if (pt.part[partnr].sysind == sysind)
				break;
		if (partnr == 4) {
			errmsg = "part: Partition not found.";
			return 0;
		}
	} else {
		errmsg = "part: Invalid argument.";
		return 0;
	}

	if (!pt.part[partnr].sysind) {
		errmsg = "part: Partition does not exist.";
		return 0;
	}

	inst->data = ud = malloc(sizeof(struct part_ud));
	ud->start = pt.part[partnr].start << 9;

	return 1;
}

static void part_remove(struct p_instance *inst)
{
	free(inst->data);
}

static int part_read(struct p_instance *inst, void *buf,
		     unsigned int off, unsigned int count)
{
	struct part_ud *ud = (struct part_ud *) inst->data;
	return p_read(inst->parent, buf, off + ud->start, count);
}

static const struct p_handles part_handles = {
	.setup  = part_setup,
	.remove = part_remove,
	.read   = part_read
};

static void part_init()
{
	p_regmod(&part_handles, "part", "Disk partitions");
}
INIT(part_init);
