/* Copyright (C) 2009-2010, Alexander Graf <agraf@znc.in>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include "cpu.h"
#include "dev.h"

/* registers */
#define DLL	(*(volatile unsigned char *)	(UART0_BASE+0x0 ))
#define DLM	(*(volatile unsigned char *)	(UART0_BASE+0x4 ))
#define THR	(*(volatile char *)		(UART0_BASE+0x0 ))
#define RBR	(*(volatile char *)		(UART0_BASE+0x0 ))
#define IER	(*(volatile int *)		(UART0_BASE+0x4 ))
#define FCR	(*(volatile unsigned char *)	(UART0_BASE+0x8 ))
#define LCR	(*(volatile unsigned char *)	(UART0_BASE+0xc ))
#define LSR	(*(volatile unsigned char *)	(UART0_BASE+0x14))

/* flags for fifo control register (fcr) */
#define FIFO_ENABLE	(1 << 0)
#define RX_RESET	(1 << 1)
#define TX_RESET	(1 << 2)

/* flags for line control register (lcr) */
#define DLAB	(1 << 7)
#define BPB(x)	(x - 5)

/* flags for line status register (lsr) */
#define RDR	(1 << 0)	/* Receiver Data Ready */
#define THRE	(1 << 5)	/* THR empty */

/* this returns the baud divisor for the specified baudrate */
#define BAUD_DIV(x)	(CPU_CLOCK / (16 * x))

void uart_init(void)
{
	/* init FIFO */
	FCR = FIFO_ENABLE | RX_RESET | TX_RESET;

	/* set baudrate */
	LCR = DLAB;
	DLL = BAUD_DIV(UART_BAUD) & 0xff;
	DLM = (BAUD_DIV(UART_BAUD) & 0xff00) >> 8;

	/* unset DLAB and set bits per byte */
	LCR = BPB(8);
}

void uart_putchar(char c)
{
	while (!(LSR & THRE)) ;

	THR = c;
}

/* wait for a char, then return it */
static char uart_getchar()
{
	while (!(LSR & RDR)) ;
	return RBR;
}

/* wait for a complete line, then return it (pointer to static buffer) */
char *uart_getline()
{
	static char linebuf[128];
	unsigned int n = 0;
	char c;

	while (n < sizeof(linebuf) - 1) {
		c = uart_getchar();
		if (c == '\n' || c == '\r') {
			uart_putchar(c);
			break;
		}
		if (c == '\b') {
			if (n > 0) {
				n--;
				/* unecho */
				uart_putchar('\b');
				uart_putchar(' ');
				uart_putchar('\b');
			}
			continue;
		}
		linebuf[n++] = c;
		uart_putchar(c);
	}

	linebuf[n] = 0;

	return linebuf;
}
